@@ -1,9 +1,13 @@
-from torch.utils.data import Dataset,DataLoader
-import math,os
+
+import random
+from torch.utils.data import Dataset, DataLoader
+import math
+import os
 import torch
 import pandas as pd
 import numpy as np
 from collections import Counter
+from typing import Callable
 
 
 class Empty(Dataset):
@@ -21,10 +25,17 @@ class Empty(Dataset):
     def __getitem__(self, idx):
         return self.seq[idx]
 
+
 class DF_Dataset(Dataset):
-    def __init__(self ,df, bs ,prepro=lambda x :x.values ,shuffle=False, col=None):
+    def __init__(
+        self, df, bs,
+        prepro=lambda x: x.values,
+        shuffle: bool = False,
+        col: str = None
+    ):
         """
-        df_dataset, a dataset for slicing pandas dataframe,instead of single indexing and collate
+        df_dataset, a dataset for slicing pandas dataframe,
+            instead of single indexing and collate
         Please use batch_size=1 for dataloader, and define the batch size here
         eg.
         ```
@@ -35,8 +46,8 @@ class DF_Dataset(Dataset):
 
         if shuffle:
             print("shuffling")
-            df = df.sample(frac=1.).reset_index().drop("index" ,axis=1)
-            print("shuffled") # shuffle the data here
+            df = df.sample(frac=1.).reset_index().drop("index", axis=1)
+            print("shuffled")  # shuffle the data here
 
         if col:
             df = df[col]
@@ -50,13 +61,14 @@ class DF_Dataset(Dataset):
     def __getitem__(self, idx):
         start = idx * self.bs
         end = (idx + 1) * self.bs
-        #         print(type(self.x_prepro(self.df[start:end])),type(self.y_prepro(self.df[start:end])))
         return self.prepro(self.df[start:end])
 
+
 class Arr_Dataset(Dataset):
     def __init__(self, *args, bs):
         """
-        arr_dataset, a dataset for slicing numpy array,instead of single indexing and collate
+        arr_dataset, a dataset for slicing numpy array,
+            instead of single indexing and collate
         Please use batch_size=1 for dataloader, and define the batch size here
         eg.
         ```
@@ -78,12 +90,24 @@ class Arr_Dataset(Dataset):
 
 class Seq_Dataset(Dataset):
     """
-    a mechanism for reading sequence data, the preparation stage of nlp learning
+    a mechanism for reading sequence data,
+        the preparation stage of nlp learning
     """
 
-    def __init__(self,
-                 seqname, seq, seq_len=None, vocab_path=None,
-                 bs=16, vocab_size=10000, build_vocab=False, sep_tok="<tok>", discard_side="right", element_type="int", fixlen=False):
+    def __init__(
+        self,
+        seqname,
+        seq,
+        seq_len: int = None,
+        vocab_path: str = None,
+        bs: int = 16,
+        vocab_size: int = 10000,
+        build_vocab: bool = False,
+        sep_tok: str = "<tok>",
+        discard_side: str = "right",
+        element_type: str = "int",
+        fixlen: bool = False
+    ):
         '''
         A pytorch dataset for sequence
         seq: enumerable sequence
@@ -103,14 +127,15 @@ class Seq_Dataset(Dataset):
         self.calc_bn()
         print(seqname, "sequence total_length type:", self.N)
 
-        self.sep_tok = sep_tok if sep_tok != None else ""
+        self.sep_tok = sep_tok if sep_tok is not None else ""
         self.make_breaker()
         self.discard_side = discard_side
         if self.seq_len:
             self.make_discard()
 
         if vocab_path:
-            if build_vocab == False and os.path.exists(vocab_path) == False:
+            if (build_vocab is False) and (
+                    os.path.exists(vocab_path) is False):
                 print("vocab path not found, building vocab path anyway")
                 build_vocab = True
             if build_vocab:
@@ -180,17 +205,22 @@ class Seq_Dataset(Dataset):
         self.process_funcs.append(self.totorch)
 
     def process_batch(self, batch):
-        return torch.nn.utils.rnn.pad_sequence(np.vectorize(self.seq2idx, otypes=[list])(batch), batch_first=True)
+        return torch.nn.utils.rnn.pad_sequence(
+            np.vectorize(self.seq2idx, otypes=[list])(batch), batch_first=True)
 
     def process_batch_fixlen(self, batch):
-        return torch.nn.utils.rnn.pad_sequence(np.vectorize(self.seq2idx, otypes=[list])(batch+[self.dummy_input()]), batch_first=True)[:-1,...]
+        return torch.nn.utils.rnn.pad_sequence(
+            np.vectorize(
+                self.seq2idx, otypes=[list])(
+                    batch+[self.dummy_input()]), batch_first=True)[:-1, ...]
 
     def dummy_input(self,):
         dummy = self.sep_tok.join([" "]*self.seq_len)
         return dummy
 
     def seq2idx(self, x):
-        for f in self.process_funcs: x = f(self, x)
+        for f in self.process_funcs:
+            x = f(self, x)
         return x
 
     def get_mapping(self, vocab_df):
@@ -213,7 +243,8 @@ class Seq_Dataset(Dataset):
 
     def get_full_token(self, list_of_tokens):
         """
-        From a list of list of tokens, to a long list of tokens, duplicate tokens included
+        From a list of list of tokens
+            to a long list of tokens, duplicate tokens included
         """
         return self.breaker(self, self.sep_tok.join(list_of_tokens))
 
@@ -224,9 +255,12 @@ class Seq_Dataset(Dataset):
         ct_dict[" "] = 7e9
         tk, ct = list(ct_dict.keys()), list(ct_dict.values())
 
-        token_df = pd.DataFrame({"token": tk, "count": ct}).sort_values(by="count", ascending=False)
-        return token_df.reset_index().drop("index", axis=1).reset_index().rename(columns={"index": "idx"}).fillna("")[
-               :self.vocab_size]
+        token_df = pd.DataFrame({"token": tk, "count": ct}).sort_values(
+            by="count", ascending=False)
+
+        return token_df.reset_index().drop("index", axis=1)\
+            .reset_index()\
+            .rename(columns={"index": "idx"}).fillna("")[:self.vocab_size]
 
 
 class fuse(Dataset):
@@ -249,6 +283,7 @@ class fuse(Dataset):
     def __getitem__(self, idx):
         return tuple(d.__getitem__(idx) for d in self.datasets)
 
+
 def collate(batch):
     return tuple(i[0] for i in zip(*batch))
 
@@ -259,7 +294,8 @@ class col_core:
         self.col_name = col_name
         self.debug = debug
         self.save_dir = save_dir
-        if self.save_dir[-1] != "/": self.save_dir += "/"
+        if self.save_dir[-1] != "/":
+            self.save_dir += "/"
 
         self.meta = dict()
 
@@ -270,18 +306,21 @@ class col_core:
         """
         pass meta dict values to obj attributes
         """
-        if meta == None: meta = self.meta
+        if meta is None:
+            meta = self.meta
         for k, v in meta.items():
-            if self.debug: print("setting:\t%s\t%s" % (k, v))
+            if self.debug:
+                print("setting:\t%s\t%s" % (k, v))
             setattr(self, k, v)
 
     def load_meta(self, path=None):
-        if path == None:
+        if path is None:
             path = self.save_dir + str(self.col_name) + ".npy"
         self.meta = np.load(path).tolist()
         self.set_meta(self.meta)
 
-        if self.meta["coltype"] == "tabulate": self.make_sub()  # make sub-objects out of meta
+        if self.meta["coltype"] == "tabulate":
+            self.make_sub()  # make sub-objects out of meta
 
     def make_meta(self):
         for attr in self.make_meta_list:
@@ -295,7 +334,9 @@ class categorical(col_core):
     def __init__(self, col_name, save_dir=".matchbox/fields"):
         super(categorical, self).__init__(col_name, save_dir)
         self.coltype = "categorical"
-        self.make_meta_list = ["col_name", "coltype", "idx2cate", "cate2idx", "width", "eye", "dim_names"]
+        self.make_meta_list = ["col_name", "coltype",
+                               "idx2cate", "cate2idx",
+                               "width", "eye", "dim_names"]
         self.__call__ = self.prepro
 
     def build(self, pandas_s, max_=20):
@@ -317,7 +358,8 @@ class categorical(col_core):
 
         self.width = len(self.cate2idx)
 
-        self.dim_names = list("%s -> %s" % (self.col_name, k) for k in self.cate2idx.keys())
+        self.dim_names = list("%s -> %s" % (self.col_name, k)
+                              for k in self.cate2idx.keys())
         self.make_meta()
 
     def trans2idx(self, cate):
@@ -337,25 +379,29 @@ class categorical(col_core):
     def __call__(self, pandas_s, expand=False):
         return self.eye[self.prepro_idx(pandas_s).values.astype(int)]
 
-    def dataset(self, df, bs,shuffle):
-        return DF_Dataset(df, prepro=self.prepro, bs=bs, shuffle=shuffle,col = self.col_name)
+    def dataset(self, df, bs, shuffle):
+        return DF_Dataset(
+            df, prepro=self.prepro, bs=bs, shuffle=shuffle, col=self.col_name)
 
 
 class categorical_idx(col_core):
     """
     preprocessor
     """
+
     def __init__(self, col_name, save_dir=".matchbox/fields"):
         """
         column name
         :param col_name: str,column name
-        :param save_dir: str, the place to save metadata, default .matchbox/fields
+        :param save_dir: str, the place to save metadata,
+            default .matchbox/fields
         """
         super(categorical_idx, self).__init__(col_name, save_dir)
         self.coltype = "categorical_idx"
         self.dim_names = [self.col_name]
         self.width = 1
-        self.make_meta_list = ["col_name", "coltype", "idx2cate", "cate2idx", "width", "dim_names"]
+        self.make_meta_list = ["col_name", "coltype",
+                               "idx2cate", "cate2idx", "width", "dim_names"]
         self.__call__ = self.prepro
 
     def build(self, pandas_s, max_=20):
@@ -384,14 +430,16 @@ class categorical_idx(col_core):
 
     def prepro(self, pandas_s, expand=True):
         x = pandas_s.apply(self.trans2idx).values
-        if expand: x = np.expand_dims(x, -1)
+        if expand:
+            x = np.expand_dims(x, -1)
         return x
 
     def __call__(self, pandas_s, expand=True):
-        return self.prepro(pandas_s,expand)
+        return self.prepro(pandas_s, expand)
 
-    def dataset(self, df, bs,shuffle):
-        return DF_Dataset(df, prepro=self.prepro, bs=bs, shuffle=shuffle,col = self.col_name)
+    def dataset(self, df, bs, shuffle):
+        return DF_Dataset(
+            df, prepro=self.prepro, bs=bs, shuffle=shuffle, col=self.col_name)
 
 
 class minmax(col_core):
@@ -402,12 +450,15 @@ class minmax(col_core):
         self.fillna = fillna
         self.dim_names = [self.col_name]
         self.width = 1
-        self.make_meta_list = ["col_name", "coltype", "min_", "max_", "range", "width", "dim_names"]
+        self.make_meta_list = ["col_name", "coltype",
+                               "min_", "max_", "range", "width", "dim_names"]
         self.__call__ = self.prepro
 
     def build(self, pandas_s=None, min_=None, max_=None):
         if type(pandas_s) != pd.core.series.Series:
-            assert (min_ != None) and (max_ != None), "If no pandas series is set you have to set min_,max_ value"
+            assert (min_ is not None) and (
+                max_ is not None),\
+                "you have to set min_,max_ value"
             self.min_ = min_
             self.max_ = max_
 
@@ -424,19 +475,25 @@ class minmax(col_core):
 
         self.range = self.max_ - self.min_
         assert self.range != 0, "the value range is 0"
-        print("min_:%.3f \tmax_:%.3f\t range:%.3f" % (self.min_, self.max_, self.range))
+        print("min_:%.3f \tmax_:%.3f\t range:%.3f" %
+              (self.min_, self.max_, self.range))
         self.make_meta()
 
     def prepro(self, data, expand=True):
-        x = (np.clip(data.values.astype(np.float64), self.min_, self.max_) - self.min_) / self.range
-        if expand: x = np.expand_dims(x, -1)
+        x = (np.clip(data.values.astype(np.float64),
+                     self.min_, self.max_) - self.min_) / self.range
+        if expand:
+            x = np.expand_dims(x, -1)
         return x
 
-    def __call__(self, data,expand=True):
-        return self.prepro(data,expand)
+    def __call__(self, data, expand=True):
+        return self.prepro(data, expand)
+
+    def dataset(self, df, bs, shuffle):
+        return DF_Dataset(
+            df, prepro=self.prepro, bs=bs,
+            shuffle=shuffle, col=self.col_name)
 
-    def dataset(self, df, bs,shuffle):
-        return DF_Dataset(df, prepro=self.prepro, bs=bs, shuffle=shuffle,col = self.col_name)
 
 class tabulate(col_core):
     def __init__(self, table_name, save_dir=".matchbox/fields"):
@@ -510,7 +567,7 @@ class tabulate(col_core):
     def __call__(self, data):
         return self.prepro(data)
 
-    def dataset(self, df, bs,shuffle):
+    def dataset(self, df, bs, shuffle):
         """
         Get the dataset from the preprocess unit
         :param df: dataframe
@@ -521,68 +578,6 @@ class tabulate(col_core):
         return DF_Dataset(df, prepro=self.prepro, bs=bs, shuffle=shuffle)
 
 
-class mapper:
-    def __init__(self, conf_path, original=None, old_default=None, new_default=None, rank_size=None):
-        """
-        Handling mapping mechanism, all index mapping should be saved as config file
-        [kwargs]
-
-        conf_path: path of the configuration file, end with npy
-        original: a list, will remove the duplicates
-        old_default: defualt original value if the interpretaion failed
-        new_default: defualt new value if the interpretaion failed
-
-        user_map = mapper("conf/user_map.npy", user_Id_List )
-        user_map.o2n will be the dictionary for original index to the new index
-        user_map.n2o will be the dictionary for new index to the original index
-
-        user_map = mappper("conf/user_map.npy") to load the conf file
-        """
-        self.conf_path = conf_path  # config file path
-        self.old_default = old_default
-        self.new_default = new_default
-        self.rank_size = rank_size
-        if original:
-            self.original = list(set(original))
-            self.mapping()
-        else:
-            self.load_map()
-
-    def mapping(self):
-        self.n2o = dict((k, v) for k, v in enumerate(self.original))
-        self.o2n = dict((v, k) for k, v in enumerate(self.original))
-        conf_dict = {"n2o": self.n2o, "o2n": self.o2n}
-        if self.rank_size:
-            chunk_size = int(len(self.original) / self.rank_size)
-            last_size = len(self.original) - (self.rank_size - 1) * chunk_size
-            conf_dict.update({"rank_len": [chunk_size] * (self.rank_size - 1) + [last_size]})
-            conf_dict.update({"rank_n2o": list(
-                dict(enumerate(self.original[chunk_size * i:chunk_size * i + conf_dict["rank_len"][i]])) for i in
-                range(self.rank_size))})
-            conf_dict.update({"rank_o2n": list(dict((v, k) for k, v in i.items()) for i in conf_dict["rank_n2o"])})
-        np.save(self.conf_path, conf_dict)
-
-    def load_map(self):
-        dicts = np.load(self.conf_path).tolist()
-        self.n2o = dicts["n2o"]
-        self.o2n = dicts["o2n"]
-        if "rank_len" in dicts:
-            self.rank_len = dicts["rank_len"]
-            self.rank_n2o = dicts["rank_n2o"]
-            self.rank_o2n = dicts["rank_o2n"]
-
-    def spit_new(self, o_idx):
-        try:
-            return self.o2n[o_idx]
-        except:
-            return self.new_default
-
-    def spit_old(self, n_idx):
-        try:
-            return self.n2o[n_idx]
-        except:
-            return self.old_default
-
 class test_DS:
     def __init__(self, dataset, *args, **kwargs):
         """
@@ -607,7 +602,8 @@ def split_df(df, valid=0.2, ensure_factor=2):
     """
     df: dataframe
     valid: valid ratio, default 0.1
-    ensure_factor, ensuring the row number to be the multiplication of this factor, default 2
+    ensure_factor, ensuring the row number
+        to be the multiplication of this factor, default 2
     return train_df, valid_df
     """
     split_ = (np.random.rand(len(df)) > valid)
@@ -617,8 +613,10 @@ def split_df(df, valid=0.2, ensure_factor=2):
     if ensure_factor:
         train_mod = len(train_df) % ensure_factor
         valid_mod = len(valid_df) % ensure_factor
-        if train_mod: train_df = train_df[:-train_mod]
-        if valid_mod: valid_df = valid_df[:-valid_mod]
+        if train_mod:
+            train_df = train_df[:-train_mod]
+        if valid_mod:
+            valid_df = valid_df[:-valid_mod]
     return train_df, valid_df
 
 
@@ -626,10 +624,11 @@ class npNormalize(object):
     """
     normalize and denormalize for numpy
     """
+
     def __init__(self, v, mean=None, std=None):
         super().__init__()
-        self.mean = v.mean() if mean == None else mean
-        self.std = v.std() if std == None else std
+        self.mean = v.mean() if mean is not None else mean
+        self.std = v.std() if std is not None else std
 
     def normalize(self, x):
         return (x - self.mean) / self.std
@@ -638,16 +637,14 @@ class npNormalize(object):
         return (x * self.std) + self.mean
 
 
-import random
-
-
 class historyReplay(object):
     """
     A historic replay scheduler for GAN training
 
     ```python
     replay = historyReplay(bs = 32, # batch size
-                current_ratio = 0.2 # for each batch keep 20% of the sample from the latest, default 0.2
+    # for each batch keep 20 persent of the sample from the latest, default 0.2
+                current_ratio = 0.2,
                 history_len = 50, # how long is the replay length, default 50
     )
     for i in range(iters):
@@ -656,6 +653,7 @@ class historyReplay(object):
         ...
     ```
     """
+
     def __init__(self, bs, current_ratio=.2, history_len=50):
         self.current_ratio = current_ratio
         self.counter = 0
@@ -676,7 +674,9 @@ class historyReplay(object):
             stack_size = self.argslist[0].size(0)
             # the 2nd ~ the history length
             if stack_size < self.bs * self.history_len:
-                self.argslist = list(torch.cat([args[i], self.argslist[i]], dim=0) for i in range(len(self.argslist)))
+                self.argslist = list(torch.cat(
+                    [args[i], self.argslist[i]], dim=0)
+                    for i in range(len(self.argslist)))
                 self.counter += 1
                 return tuple(args) if self.arglen > 1 else tuple(args)[0]
             # above history length
@@ -684,10 +684,67 @@ class historyReplay(object):
                 pos = self.counter % self.history_len
                 start_pos = pos * self.bs
                 end_pos = (pos + 1) * self.bs
-                slice_ = random.choices(range(self.bs * self.history_len), k=self.history_chunk)
+                slice_ = random.choices(
+                    range(self.bs * self.history_len), k=self.history_chunk)
                 rt = []
                 for i in range(len(self.argslist)):
-                    rt.append(torch.cat([args[i][:self.latest_chunk, ...], self.argslist[i][slice_, ...]], dim=0))
+                    rt.append(torch.cat(
+                        [args[i][:self.latest_chunk, ...],
+                         self.argslist[i][slice_, ...]], dim=0))
                     self.argslist[i][start_pos:end_pos, ...] = args[i]
                 self.counter += 1
-                return tuple(rt) if self.arglen > 1 else tuple(rt)[0]
\ No newline at end of file
+                return tuple(rt) if self.arglen > 1 else tuple(rt)[0]
+
+
+def layering(dataset_class, new_name):
+    """
+    Instead of creating a bunch of dataset
+    just treat another dataset as extra layer of __getitem__ function
+
+    Inputs:
+        - dataset_class: Dataset class, or functions
+            that are decorated with this function
+        - new_name: a name for next layer class
+
+    eg.
+
+    @layering(SomeStringDataset, "tensorDataset")
+    def tokenizing(x):
+        return tokenizer(x, return_tensors='pt')['input_ids'][0]
+
+    @layering(tokenizing, "guessNextDataset")
+    def guess_next(x):
+        return x[:-1], x[1:]
+
+    some_string_data = SomeStringDataset()
+
+    tokenized_data = some_string_data.next_layer()
+    print(tokenized_data[3])
+
+    guess_pair = tokenized_data.next_layer()
+    print(guess_pair[3])
+    """
+    if hasattr(dataset_class, "associated_class"):
+        dataset_class = dataset_class.associated_class
+
+    def decorator(f: Callable):
+        class newClass(Dataset):
+            def __init__(self, last_ds,):
+                self.last_ds = last_ds
+
+            def __repr__(self):
+                return f"Layered Pytorch Dataset: {new_name}"
+
+            def __len__(self): return len(self.last_ds)
+
+            def __getitem__(self, idx):
+                return f(self.last_ds[idx])
+
+        newClass.__name__ = new_name
+
+        def next_layer(self,):
+            return newClass(self)
+        f.associated_class = newClass
+        dataset_class.next_layer = next_layer
+        return f
+    return decorator
